#!/usr/bin/perl

# /usr/local/bin/sensor.pl
# Dieses Programm liest die Sensor- und Alarmdaten vom Sensormodul 2 oder 3 über die RS-232-Schnittstelle oder USB ein
# und speichert die aufbereiteten Daten sowie ein Alarm-Log. Über die Datei map kann die Sensorzuordnung eingestellt
# werden. Weiterhin werden ausgewählte Daten an den MQTT-Server Mosquitto übermittelt, sofern dieser installiert ist.
# Scott-Falk Hühn, 28.12.2021

use strict;
use warnings;
use Net::MQTT::Simple;

my $senif = "/dev/serial0";            # Serielle Schnittstelle des Raspberry Pi
#my $senif = "/dev/ttyUSB0";            # USB wird anstatt der seriellen Schnittstelle verwendet
my $sendir = "/mnt/ram";               # Verzeichnis für Sensor- und Alarmdaten (flüchtige RAM-Disk)
my $cfgdir = "/srv/sensor";            # Verzeichnis für Sensorkonfiguration, Alarm-Log und Sensorbelegung
my $mqtt_inst = "/usr/sbin/mosquitto"; # Pfad zum Programm Mosquitto, wird zur Installationsprüfung benötigt
my $mqtt_serv = "localhost";           # IP-Adresse des MQTT-Servers, z.B. 192.168.1.122 (oder localhost)
my $mqtt_port = "1883";                # Port des MQTT-Servers, normal 1883
my $mqtt_ok = 0;                       # Flag für installierten MQTT-Server Mosquitto
my $mqtt;                              # Verbindung zum MQTT-Server
my $SER;                               # Dateihandle der seriellen Schnittstelle
my $AL;                                # Dateihandle einer Alarmdatei
my $SE;                                # Dateihandle einer Sensordatei
my $ALOG;                              # Dateihandle des Alarm-Logs
my $MAP;                               # Dateihandle einer Map-Datei (Sensorzuordnung)
my $CFG;                               # Dateihandle einer Konfigurationsdatei
my @alist;                             # Zwischenspeicher für Alarm-Log
my @mlist;                             # Zwischenspeicher für Map-Datei
my @talist;                            # MQTT-Topic-Liste für Alarme
my @ralist;                            # Ruhezustand-Liste der Alarme
my @tslist;                            # MQTT-Topic-Liste für Sensoren
my $mflag;                             # Flag für Map-Modus (1 = Map aktiv)
my $alnum;                             # empfangene Alarmnummer
my $alval;                             # empfangener Alarmwert
my $alidl;                             # Alarm-Ruhezustand
my $alold;                             # letzter gespeicherter Alarmwert
my $almax;                             # maximale Einträge im Alarm-Log
my $senum;                             # empfangene Sensornummer
my $seval;                             # empfangener Sensorwert
my $secnt;                             # Sensorzähler für Sensorbelegung
my $sebus;                             # Sensoranzahl an einem Bus
my $topic;                             # Zwischenspeicher für MQTT-Topic
my $min;                               # aktuelle Minute
my $i;                                 # Zählervariable
my $height;                            # Höhe für Luftdruckberechnung
my $test;                              # Flag für Test-Modus, wird durch Parameter "t" oder "T" beim Start aktiviert

$test = $ARGV[0];                                             # Paramter 1 holen
if (defined $test) {                                          # wenn Parameter 1 angegeben
  if (($test eq 't') || ($test eq 'T')) { $test = 1; }        # wenn Zeichen 't' oder 'T' -> Test einschalten
  else { $test = 0; }                                         # sonst Test ausschalten
}

if (-e $mqtt_inst) {                                          # wenn MQTT-Server Mosquitto installiert ist
  $mqtt_ok = 1;                                               # Flag setzen
  $mqtt = Net::MQTT::Simple->new("$mqtt_serv:$mqtt_port");    # Verbindung zum MQTT-Server herstellen
}
if ($test) {                                                  # wenn Test-Modus aktiv
  print "MQTT-Verbindung ";                                   # MQTT-Status ausgeben
  if ($mqtt_ok == 1) { print "aktiv\n" }
  else { print "nicht aktiv\n" }
}

system("stty -F $senif 9600 -echo");                          # Baudrate einstellen und Echo ausschalten
$mflag = 0;                                                   # Flag für Map-Modus löschen
if (open($MAP, "<$cfgdir/map")) {                             # wenn Zuordnungsdatei vorhanden -> öffnen
  @mlist = <$MAP>;                                            # komplette Zuordnungsdatei einlesen
  close($MAP);                                                # Zuordnungsdatei schließen
  $mflag = 1;                                                 # Flag für Map-Modus setzen
}
read_cfg();                                                   # Konfigurationsdateien einlesen
if ($test) { print "Konfiguration gelesen\n"; }               # wenn Test-Modus aktiv -> 

while (1) {                                                   # Endlosschleife
  open($SER, "<$senif") or die $!;                            # serielle Schnittstelle öffnen
  while (<$SER>) {                                            # wenn ein Datenpaket empfangen wurde
    chomp($_);                                                # Steuerzeichen entfernen
    if (length($_) == 3 and substr($_, 1, 1) eq ":") {        # wenn Alarmmeldung empfangen wurde
      $alnum = ord($_) - ord("A") + 1;                        # Alarmnummer ermitteln
      $alval = ord(substr($_, 2, 1)) - ord("0");              # Alarmwert ermitteln
      if ($alnum >= 1 and $alnum <= 4) {                      # wenn gültige Alarmnummer
        if ($alval >= 0 and $alval <= 1) {                    # wenn gültiger Alarmwert
          if (open($AL, "<$sendir/al$alnum")) {               # wenn Alarmdatei vorhanden -> öffnen
            $alold = <$AL>;                                   # bisherigen Alarmstatus lesen
            close($AL);                                       # Alarmdatei schließen
            chomp($alold);                                    # Zeilenendezeichen entfernen
          }
          else { $alold = $alval }                            # keine Alarmdatei -> kein Statuswechsel
          open($AL, ">$sendir/al$alnum");                     # neue Alarmdatei schreiben
          print $AL "$alval\n";                               # neuen Alarmstatus schreiben
          close($AL);                                         # Alarmdatei schließen
          if ($test) { print "a$alnum:$alval\n" }             # wenn Test-Modus aktiv -> Alarm ausgeben
          $topic = $talist[$alnum - 1];                       # Topic-Eintrag aus Liste holen
          $alidl = $ralist[$alnum - 1];                       # Alarm-Ruhezustand aus Liste holen
          if ($topic gt '') {                                 # wenn Topic vorhanden
            if ($mqtt_ok) {                                   # wenn MQTT-Server vorhanden
              if ($alidl == $alval) { $alidl = 0; }           # wenn aktueller Zustand = Ruhezustand -> 0 setzen
              else { $alidl = 1; }                            # sonst 1 setzen
              $mqtt->publish("$topic" => "$alidl");           # MQTT-Publish von Topic und Alarmwert
            }
          }
          if ($alval != $alold) {                             # wenn Alarmstatuswechsel
            if (open($ALOG, "<$cfgdir/alog")) {               # wenn Alarm-Log vorhanden -> öffnen
              @alist = <$ALOG>;                               # komplettes Log einlesen
              close($ALOG);                                   # Alarm-Log schließen
            }
            else { @alist = () }                              # kein Alarm-Log -> leeres Log anlegen
            open($ALOG, ">$cfgdir/alog");                     # Alarm-Log schreiben
            my ($sec,$min,$hou,$day,$mon,$yea,$wday,$yday,$dst) = localtime(time); # aktuelle Zeit
            if ($sec < 10) { $sec = "0".$sec }                # Sekunden einstellig -> 0 voranstellen
            if ($min < 10) { $min = "0".$min }                # Minuten einstellig -> 0 voranstellen
            if ($hou < 10) { $hou = "0".$hou }                # Stunden einstellig -> 0 voranstellen
            if ($day < 10) { $day = "0".$day }                # Tag einstellig -> 0 voranstellen
            $mon += 1;                                        # Monat korrigieren (0-11 -> 1-12)
            if ($mon < 10) { $mon = "0".$mon }                # Monat einstellig -> 0 voranstellen
            print $ALOG $day . "." . $mon . "." . ($yea + 1900) . " ";  # neuer Eintrag, Datum schreiben
            print $ALOG $hou . ":" . $min . ":" . $sec . " " . $alnum . $alval . "\n"; # Zeit und Alarmdaten
            $almax = scalar(@alist);                          # Anzahl der Einträger in der Alarmliste
            if ($almax > 99) { $almax = 99 }                  # Anzahl auf 99 begrenzen
            for ($i = 0; $i < $almax; $i++) {                 # Einträge bis zur ermittelten Länge bearbeiten
              print $ALOG $alist[$i];                         # Einträge in Alarm-Log schreiben
            }
            close($ALOG);                                     # Alarm-Log schließen
          }
        }
      }
    }
    if (length($_) == 7 and substr($_, 1, 1) eq ":") {        # wenn Sensorwert empfangen wurde
      $senum = ord($_) - ord("0");                            # Sensornummer ermitteln
      if ($senum > 8) {                                       # wenn Sensor A-P
        $senum = $senum - 40;                                 # Sensornummer korrigieren
      }
      $seval = substr($_, 2, 5);                              # Sensorwert ermitteln
      if ($senum >= 1 and $senum <= 27) {                     # wenn Temperatursensor
        if ($seval =~ /^\s*\-?\d*\.?\d?$/) {                  # gültige Temperatur?
          if ($seval >= -55 and $seval <= 125) {              # gültiger Wertebereich?
            if ($mflag) {                                     # wenn Map-Flag gesetzt
              $senum = $mlist[$senum - 1];                    # zugeordnete Sensornummer holen
              chomp($senum);                                  # Steuerzeichen entfernenn
            }
            open($SE, ">$sendir/s$senum");                    # Sensordatei anlegen
            print $SE "$seval\n";                             # Sensorwert schreiben
            close($SE);                                       # Sensordatei schließen
            if ($test) { print "s$senum:$seval\n" }           # wenn Test-Modus aktiv -> Sensordaten ausgeben
            $topic = $tslist[$senum - 1];                     # Topic-Eintrag aus Liste holen
            if ($topic gt '') {                               # wenn Topic vorhanden
              if ($mqtt_ok) {                                 # wenn MQTT-Server vorhanden
                $seval =~ s/^\s+|\s+$//g;                     # Leerzeichen entfernen
                $mqtt->publish("$topic" => "$seval");         # MQTT-Publish von Topic und Sensorwert
              }
            }
          }
        }
      }
      if ($senum == 31) {                                     # wenn Luftdrucksensor
        if ($seval =~ /^\s*\d*$/) {                           # gültiger Luftdruck?
          if ($seval >= 200 and $seval <= 1100) {             # gültiger Wertebereich?

            # Der Luftdruckwert wird jetzt auf Meereshöhe umgerechnet und auf ganze Zahlen gerundet, die
            # die barometrische Formel stammt von: https://rechneronline.de/barometer
            # Luftdruck auf Meereshöhe = Barometeranzeige / (1 - Temperaturgradient * Höhe / Temperatur auf Meereshöhe
            # in Kelvin) ^ (0.03416 / Temperaturgradient)              (Temperaturgradient = 0.0065)

            $seval = sprintf("%5.0f", $seval / (1 - 0.0065 * $height / 288.15) ** (0.03416 / 0.0065));
            open($SE, ">$sendir/s28");                        # Sensordatei anlegen
            print $SE "$seval\n";                             # Sensorwert schreiben
            close($SE);                                       # Sensordatei schließen
            if ($test) { print "s28:$seval\n" }               # wenn Test-Modus aktiv -> Sensordaten ausgeben
            $topic = $tslist[27];                             # Topic-Eintrag aus Liste holen
            if ($topic gt '') {                               # wenn Topic vorhanden
              if ($mqtt_ok) {                                 # wenn MQTT-Server vorhanden
                $seval =~ s/^\s+|\s+$//g;                     # Leerzeichen entfernen
                $mqtt->publish("$topic" => "$seval");         # MQTT-Publish von Topic und Sensorwert
              }
            }
          }
        }
      }
      if ($senum >= 32 and $senum <= 34) {                    # wenn Luftfeuchtigkeitssensor
        if ($seval =~ /^\s*\d*$/) {                           # gültige Luftfeuchtigkeit?
          if ($seval >= 0 and $seval <= 100) {                # gültiger Wertebereich?
            if ($mflag) {                                     # wenn Map-Flag gesetzt
              $senum = $mlist[$senum - 8] + 7;                # zugeordnete Sensornummer holen
              chomp($senum);                                  # Steuerzeichen entfernen
            }
            open($SE, ">$sendir/s" . ($senum - 3));           # Sensordatei anlegen
            print $SE "$seval\n";                             # Sensorwert schreiben
            close($SE);                                       # Sensordatei schließen
            if ($test) { print "s$senum:$seval\n" }           # wenn Test-Modus aktiv -> Sensordaten ausgeben
            $topic = $tslist[$senum - 4];                     # Topic-Eintrag aus Liste holen
            if ($topic gt '') {                               # wenn Topic vorhanden
              if ($mqtt_ok) {                                 # wenn MQTT-Server vorhanden
                $seval =~ s/^\s+|\s+$//g;                     # Leerzeichen entfernen
                $mqtt->publish("$topic" => "$seval");         # MQTT-Publish von Topic und Sensorwert
              }
            }
          }
        }
      }
    }
    if (length($_) == 8) {                                    # wenn Sensorbelegung empfangen wurde
      if ($_ =~ /^[0-9A-R]*$/) {                              # gültige Sensorbelegung?
        $secnt = 0;                                           # Sensor-Gesamtanzahl = 0 setzen
        for ($i = 0; $i < 8; $i++) {                          # 8 Sensorbusse prüfen
          $sebus = ord(substr($_, $i, 1)) - 48;               # Sensoranzahl am Bus ermitteln
          if ($sebus > 9) {                                   # wenn Buchstabe
            $sebus -= 7;                                      # in entsprechende Zahl wandeln
          }
          $secnt += $sebus;                                   # Sensoren am Bus zur Gesamtanzahl addieren
        }
        open($SE, ">$cfgdir/sen");                            # Sensorbelegungsdatei anlegen
        print $SE "$_ ($secnt)\n";                            # Sensorbelegung und Gesamtanzahl schreiben
        close($SE);                                           # Sensorbelegungsdatei schließen
        if ($test) { print "sen:$_ ($secnt)\n" }              # wenn Test-Modus aktiv -> Sensorbelegung ausgeben
      }
    }
    if (-e "$cfgdir/change") {                                # wenn Change-Datei vorhanden
      unlink "$cfgdir/change";                                # Change-Datei wieder löschen
      read_cfg();                                             # Konfigurationsdateien neu einlesen
    }
  }
  close($SER);                                                # Schleife neu starten, falls EOT (4) empfangen wird
}

sub read_cfg {                                                # Unterprogramm zum Lesen der Konfigurationsdateien
  undef @tslist;                                              # aktuelle Sensor-Topic-Liste löschen
  for ($i = 1; $i < 32; $i ++) {                              # Konfiguration von 31 Sensoren lesen
    if (open($CFG, "<$cfgdir/s$i")) {                         # wenn Konfigurationsdatei vorhanden -> öffnen
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 1 lesen und verwerfen (Sensorname)
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 2 lesen und verwerfen (unterer Grenzwert)
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 3 lesen und verwerfen (oberer Grenzwert)
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 4 lesen und verwerfen (Sperre)
      if (!eof($CFG)) {                                       # wenn kein Dateiende
        $topic = <$CFG>;                                      # Zeile 5 lesen (MQTT-Topic)
        chomp($topic);                                        # Zeilenendezeichen entfernen
      }
      else { $topic = ''; }                                   # sonst leere Zeile setzen
      close($CFG);                                            # Konfigurationsdatei schließen
    }
    else { $topic = ''; }                                     # wenn keine Konfigurationsdatei -> leere Zeile setzen
    push @tslist, $topic;                                     # MQTT-Topic an Liste anfügen
  }
  undef @talist;                                              # aktuelle Alarm-Topic-Liste löschen
  undef @ralist;                                              # aktuelle Alarm-Ruhezustand-Liste löschen
  for ($i = 1; $i < 5; $i ++) {                               # Konfiguration von 4 Alarmen lesen
    if (open($CFG, "<$cfgdir/al$i")) {                        # wenn Konfigurationsdatei vorhanden -> öffnen
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 1 lesen und verwerfen (Alarmname)
      if (!eof($CFG)) {                                       # wenn kein Dateiende
        $alidl = <$CFG>;                                      # Zeile 2 lesen (Ruhezustand)
        chomp($alidl);                                        # Zeilenendezeichen entfernen
      }
      else { $alidl = '0'; }                                  # sonst Ruhezustand 0 setzen
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 3 lesen und verwerfen (Normalzustand)
      if (!eof($CFG)) { $topic = <$CFG>; }                    # Zeile 4 lesen und verwerfen (Alarmzustand)
      if (!eof($CFG)) {                                       # wenn kein Dateiende
        $topic = <$CFG>;                                      # Zeile 5 lesen (MQTT-Topic)
        chomp($topic);                                        # Zeilenendezeichen entfernen
      }
      else { $topic = ''; }                                   # sonst leere Zeile setzen
      close($CFG);                                            # Konfigurationsdatei schließen
    }
    else {                                                    # wenn keine Konfigurationsdatei
      $alidl = '0';                                           # Ruhezustand 0 setzen
      $topic = '';                                            # leere Zeile setzen
    }
    push @talist, $topic;                                     # MQTT-Topic an Liste anfügen
    push @ralist, $alidl;                                     # Alarm-Ruhezustand an Liste anfügen
  }
  $height = 0;                                                # Höhe auf 0 voreinstellen
  if (open($CFG, "<$cfgdir/height")) {                        # wenn Höhendatei vorhanden -> öffnen
    $height = <$CFG>;                                         # Höhe lesen
    chomp($height);                                           # Zeilenendezeichen entfernen
    close($CFG);                                              # Höhendatei schließen
  }
}
